unit CaveGraph;

interface

uses Classes,Vector,DSupport,HugeArry,Hash,CaveTypes,CaveLib;


type TAdjacencyList = class(THugeArray)
 private
  procedure Put(Index: integer; Item: THugeInteger);
  function Get(Index: integer): THugeInteger;
 protected
 public
  constructor Create(Number: integer);
  destructor Destroy; override;
  function Add: THugeInteger;
  procedure Delete(Index: integer);
  property Value[Index: integer]: THugeInteger read Get; default;
  procedure Clear; override;
 end;


type TEdge = record
 Inx1,Inx2,Color: integer;
 end;


type TEdgeArray = class(TObject)
 private
  Edges1: THugeInteger;
  Edges2: THugeInteger;
  function GetEdge(Inx: integer): TEdge;
  procedure SetEdge(Inx: integer; Edge: TEdge);
  function GetItemIndex: integer;
  procedure SetItemIndex(Value: integer);
 public
  Common,SecondHalf,EdgeValue: integer;
  Color: THugeInteger;
  constructor Create;
  destructor Destroy; override;
  function Add(Edge: TEdge): integer;
  procedure Clear;
  function Optimise(First,Last: integer): integer;
  property Edges[Inx: integer]: TEdge read GetEdge write SetEdge; default;
  property ItemIndex: integer read GetItemIndex write SetItemIndex;
  procedure Delete(Index: integer);
  procedure Assign(Source: TEdgeArray);
  function FindEdge(E: TEdge): integer;
  procedure SetColors(V: integer);
 end;




type TSurveyGraph = class(TObject)
 private
 public
  Names: THashTable;
  Vertices: THuge3DVector;
  VertexValue: THugeInteger;
  AdjacentVertices: TAdjacencyList;
  AdjacentEdges: TAdjacencyList;
  Parent: THugeInteger;
  Edges: TEdgeArray;
  BoundingBox: TBoundBox;
  constructor Create;
  destructor Destroy; override;
  procedure Clear;
  function AddVertex(Name: shortstring; Loc: T3DVector): integer;
  procedure GetVertex(Inx: integer; var Name: shortstring; var Loc: T3DVector);
  function AddEdge(Station1,Station2: shortstring): boolean; overload;
  procedure AddEdge(Inx1,Inx2: integer); overload;
  function AddEdge(Edge: TEdge): boolean; overload;
  procedure GetEdge(Inx: integer; var Name1,Name2: shortstring; var V1,V2: T3DVector);
  procedure SetColors(V: integer);
 end;



type TLoopList = class(THugeArray)
 private
  procedure Put(Index: integer; Item: TEdgeArray);
  function Get(Index: integer): TEdgeArray;
 protected
 public
  constructor Create(Number: integer);
  destructor Destroy; override;
  function Add: TEdgeArray;
  procedure Delete(Index: integer);
  property Value[Index: integer]: TEdgeArray read Get; default;
  procedure Clear; override;
 end;

function EqualEdges(E1,E2: TEdge): boolean;
procedure CopyAllVertices(Source,Destination: TSurveyGraph);


implementation


function SortEdges(Edges: TEdgeArray; L, R: Integer): integer;
{Optimise edge list by sorting according the second index in edge}
var I, J: Integer;
var P, T: TEdge;
begin
Result:=0;
repeat
	begin
	I:= L; J:= R;
	P:= Edges[(L + R) shr 1];
	repeat
		begin
		while Edges[I].Inx2 < P.Inx2 do Inc(I);
		while Edges[J].Inx2 > P.Inx2 do Dec(J);
		if I <= J then
			begin
			T:= Edges[I]; Edges[I]:= Edges[J]; Edges[J]:= T;
			Inc(I); Dec(J);
			Inc(Result);
			end;
		end
	until I > J;
	if L < J then Result:=Result+SortEdges(Edges, L, J);
	L:= I;
	end
until I >= R;
end;



function OptimiseEdges(EdgeArray: TEdgeArray; First,Last: integer): integer;
{Optimise sequence of edges between start and stop}
var I: integer;
var T: integer;
var Edge: TEdge;
begin
Result:=0;
for I:=First to Last do
	begin
        Edge:=EdgeArray[I];
	if SwapIntegerArgs(Edge.Inx1>Edge.Inx2,Edge.Inx1,Edge.Inx2) then
		begin
		EdgeArray[I]:=Edge;
		Inc(Result);
		end;
	end;
SortEdges(EdgeArray,First,Last);
end;



{-------------------------------------------------------------------------}
{ Implementation For Edge Array Object                                    }
{-------------------------------------------------------------------------}


function TEdgeArray.Optimise(First,Last: integer): integer;
begin
Result:=OptimiseEdges(Self,First,Last);
end;


function TEdgeArray.GetEdge(Inx: integer): TEdge;
begin
Result.Inx1:=Edges1[Inx];
Result.Inx2:=Edges2[Inx];
Result.Color:=Color[Inx];
end;


procedure TEdgeArray.SetEdge(Inx: integer; Edge: TEdge);
begin
Edges1.Value[Inx]:=Edge.Inx1;
Edges2.Value[Inx]:=Edge.Inx2;
Color[Inx]:=Edge.Color;
end;


function TEdgeArray.GetItemIndex: integer;
begin
Result:=Edges1.ItemIndex;
end;


procedure TEdgeArray.SetItemIndex(Value: integer);
begin
if Value<>Edges1.ItemIndex then
	begin
	Edges1.ItemIndex:=Value;
	Edges2.ItemIndex:=Value;
	Color.ItemIndex:=Value;
	end;
end;


procedure TEdgeArray.SetColors(V: integer);
var I: integer;
begin
for I:=0 to Color.ItemIndex-1 do Color[I]:=V;
end;


procedure TEdgeArray.Clear;
begin
Edges1.Clear;
Edges2.Clear;
Color.Clear;
end;


procedure TEdgeArray.Delete(Index: integer);
begin
Edges1.Delete(Index);
Edges2.Delete(Index);
Color.Delete(Index);
end;



function TEdgeArray.Add(Edge: TEdge): integer;
begin
Result:=Edges1.Add(Edge.Inx1);
Edges2.Add(Edge.Inx2);
Color.Add(0);
end;


function TEdgeArray.FindEdge(E: TEdge): integer;
begin
for Result:=0 to Edges1.ItemIndex-1 do
	begin
	if (E.Inx1=Edges1[Result]) and (E.Inx2=Edges2[Result]) then exit;
	if (E.Inx1=Edges2[Result]) and (E.Inx2=Edges1[Result]) then exit;
	end;
Result:=-1;
end;



procedure TEdgeArray.Assign(Source: TEdgeArray);
var I: integer;
begin
Clear;
Self.Common:=Source.Common;
Self.SecondHalf:=Source.SecondHalf;
Self.EdgeValue:=Source.EdgeValue;
for I:=0 to TEdgeArray(Source).ItemIndex-1 do Add(Source[I]);
end;



constructor TEdgeArray.Create;
begin
Edges1:=THugeInteger.Create(5);
Edges2:=THugeInteger.Create(5);
Color:=THugeInteger.Create(5);
end;



destructor TEdgeArray.Destroy;
begin
Edges1.Free;
Edges2.Free;
Color.Free;
inherited Destroy;
end;


{-------------------------------------------------------------------------}
{ Implementation Adjacency List object based on huge array object         }
{-------------------------------------------------------------------------}


constructor TAdjacencyList.Create(Number: Integer);
{Create a huge Adjacency List}
begin
inherited Create(1,SizeOf(THugeInteger));
Add;
end;


function TAdjacencyList.Add: THugeInteger;
{Add item to Adjacency List}
var PTR: ^THugeInteger;
begin
Result:=THugeInteger.Create(5);
ExpandArray;
PTR:=GetItemPtr(ItemIndex);
PTR^:=Result;
ItemIndex:=ItemIndex+1;
end;


procedure TAdjacencyList.Put(Index: integer; Item: THugeInteger);
{Insert Adjacency List at index}
var PTR: ^THugeInteger;
begin
PTR:=GetItemPtr(Index);
PTR^:=Item;
end;


function TAdjacencyList.Get(Index: Longint): THugeInteger;
{Get Adjacency List at the index}
var PTR: ^THugeInteger;
begin
PTR:=GetItemPtr(Index);
Result:=PTR^;
end;



procedure TAdjacencyList.Delete(Index: integer);
begin
Get(Index).Free;
DeleteItem(Index);
end;


procedure TAdjacencyList.Clear;
var I: integer;
begin
for I:=0 to ItemIndex-1 do Get(I).Free;
inherited Clear;
end;


destructor TAdjacencyList.Destroy;
{Destroy Huge array item}
var I: integer;
begin
Clear;
inherited Destroy;
end;


{-------------------------------------------------------------------------}
{ Implementation of station array Object                                  }
{-------------------------------------------------------------------------}



procedure TSurveyGraph.Clear;
begin
Names.Strings.ItemIndex:=0;
Names.Rehash;
Vertices.Clear;
VertexValue.Clear;
Parent.Clear;
Edges.Clear;
AdjacentVertices.Clear;
AdjacentEdges.Clear;
BoundingBox.Max.Vector:=MakeVector3D(-9.9e99,-9.9e99,-9.9e99);
BoundingBox.Min.Vector:=MakeVector3D(9.9e99,9.9e99,9.9e99);
end;


function TSurveyGraph.AddVertex(Name: shortstring; Loc: T3DVector): integer;
begin
Names.Insert(Name);
Result:=Vertices.Add(Loc);
VertexValue.Add(0);
Parent.Add(-1);
AdjacentVertices.Add;
AdjacentEdges.Add;
if Result = 0 then
	begin
	BoundingBox.Max.Vector:=Loc;
	BoundingBox.Min.Vector:=Loc;
	end
else
	begin
	BoundingBox.Max.Vector:=GetMaximumVector(BoundingBox.Max.Vector,Loc);
	BoundingBox.Min.Vector:=GetMinimumVector(BoundingBox.Min.Vector,Loc);
	end;
end;



procedure TSurveyGraph.GetVertex(Inx: integer; var Name: shortstring; var Loc: T3DVector);
begin
Name:=Names[Inx];
Loc:=Vertices[Inx];
end;


function TSurveyGraph.AddEdge(Edge: TEdge): boolean;
var Inx: integer;
begin
Result:=False;
if (Edge.Inx1<0) or (Edge.Inx2<0) then exit;
Inx:=Edges.Add(Edge);
AdjacentVertices.Value[Edge.Inx1].Add(Edge.Inx2);
AdjacentVertices.Value[Edge.Inx2].Add(Edge.Inx1);
AdjacentEdges.Value[Edge.Inx1].Add(Inx);
AdjacentEdges.Value[Edge.Inx2].Add(Inx);
Parent[Edge.Inx2]:=Edge.Inx1;
Result:=True;
end;


procedure TSurveyGraph.AddEdge(Inx1,Inx2: integer);
var Edge: TEdge;
begin
Edge.Inx1:=Inx1;
Edge.Inx2:=Inx2;
AddEdge(Edge);
end;





function TSurveyGraph.AddEdge(Station1,Station2: shortstring): boolean;
var Edge: TEdge;
begin
Result:=False;
Edge.Inx1:=Names.Search(Station1);
if Edge.Inx1<0 then exit;
Edge.Inx2:=Names.Search(Station2);
Result:=AddEdge(Edge);
end;



procedure TSurveyGraph.GetEdge(Inx: integer; var Name1,Name2: shortstring; var V1,V2: T3DVector);
var Edge: TEdge;
begin
Edge:=Edges[Inx];
Name1:=Names.Strings.Value[Edge.Inx1];
Name2:=Names.Strings.Value[Edge.Inx2];
V1:=Vertices[Edge.Inx1];
V2:=Vertices[Edge.Inx2];
end;



procedure TSurveyGraph.SetColors(V: integer);
var I: integer;
begin
for I:=0 to VertexValue.ItemIndex-1 do VertexValue[I]:=V;
end;


constructor TSurveyGraph.Create;
begin
inherited Create;
Names:=THashTable.Create(5,StationNameSize);
Vertices:=THuge3DVector.Create(5);
VertexValue:=THugeInteger.Create(5);
Parent:=THugeInteger.Create(5);
Edges:=TEdgeArray.Create;
AdjacentVertices:=TAdjacencyList.Create(5);
AdjacentEdges:=TAdjacencyList.Create(5);
end;



destructor TSurveyGraph.Destroy;
begin
Names.Free;
Vertices.Free;
VertexValue.Free;
Parent.Free;
Edges.Free;
AdjacentVertices.Free;
AdjacentEdges.Free;
inherited Destroy;
end;




{-------------------------------------------------------------------------}
{ Implementation Adjacency List object based on huge array object         }
{-------------------------------------------------------------------------}


constructor TLoopList.Create(Number: Integer);
{Create a huge Adjacency List}
begin
inherited Create(Number,SizeOf(TEdgeArray));
end;


function TLoopList.Add: TEdgeArray;
{Add item to Adjacency List}
var PTR: ^TEdgeArray;
begin
Result:=TEdgeArray.Create;
ExpandArray;
PTR:=GetItemPtr(ItemIndex);
PTR^:=Result;
ItemIndex:=ItemIndex+1;
end;


procedure TLoopList.Put(Index: integer; Item: TEdgeArray);
{Insert Adjacency List at index}
var PTR: ^TEdgeArray;
begin
PTR:=GetItemPtr(Index);
PTR^:=Item;
end;


function TLoopList.Get(Index: Longint): TEdgeArray;
{Get Adjacency List at the index}
var PTR: ^TEdgeArray;
begin
PTR:=GetItemPtr(Index);
Result:=PTR^;
end;



procedure TLoopList.Delete(Index: integer);
begin
Get(Index).Free;
DeleteItem(Index);
end;


procedure TLoopList.Clear;
var I: integer;
begin
for I:=0 to ItemIndex-1 do Get(I).Free;
inherited Clear;
end;



destructor TLoopList.Destroy;
{Destroy Huge array item}
begin
Clear;
inherited Destroy;
end;


{---------------------- Support Routines --------------------------------------}


function EqualEdges(E1,E2: TEdge): boolean;
begin
Result:=((E1.Inx1 = E2.Inx1) and (E1.Inx2 = E2.Inx2)) or
        ((E1.Inx1 = E2.Inx2) and (E1.Inx2 = E2.Inx1));
end;




procedure CopyAllVertices(Source,Destination: TSurveyGraph);
var I: integer;
var S: shortstring;
var V: T3DVector;
begin
Destination.Clear;
for I:=0 to Source.Vertices.ItemIndex-1 do
	begin
	Source.GetVertex(I,S,V);
	Destination.AddVertex(S,V);
	end;
end;




end.
