unit CaveLoops;

{Unit to find loops in Cave Survey Graphs }

interface

uses Classes,Vector,HugeArry,CaveGraph;


procedure FindSpanningTree(Source,Destination: TSurveyGraph; CloseEdges: TLoopList; ByIncidence: boolean);
procedure FindLoopEdges(SpanTree: TSurveyGraph; EdgeArray: TEdgeArray);
procedure FindLoops(Source: TSurveyGraph; LoopList: TLoopList; Optimize: boolean);


implementation

uses LoopMain;


{------------- Heap Object for Generating BFS Spanning Tree -------------------}

{ Object keeps track of adjacent vertexes at each level of the BFS traverse.}

type TSpanningHeap = class(TObject)
 public
  ParentInx: THugeInteger;
  ChildInx: THugeInteger;
  Lengths: THugeDouble;
  AdjCounts: THugeInteger;
  procedure Add(Parent,Child: integer; ParentVertex,ChildVertex: T3DVector; AdjCount: integer);
  function GetBestHeapItem(ByIncidence: boolean): integer;
  procedure Delele(Inx: integer);
  constructor Create;
  destructor Destroy; override;
  end;



procedure TSpanningHeap.Delele(Inx: integer);
{Delete specified heap item}
begin
ParentInx.Delete(Inx);
ChildInx.Delete(Inx);
Lengths.Delete(Inx);
AdjCounts.Delete(Inx);
end;



function TSpanningHeap.GetBestHeapItem(ByIncidence: boolean): integer;
{Find 'Best' vertex on the heap. If "ByIncidence" is true}
{returns the vertex with the most connection. Otherwise it}
{returns the vertext with shortest connection. (Prim's algorithm)}
{Could be optimized using sorted lists or a binary tree}
var I: integer;
begin
Result:=0;
if ByIncidence then
	begin
	{Find vertex with the maximum connections}
	for I:=0 to AdjCounts.ItemIndex-1 do
	 if AdjCounts[I]>AdjCounts[Result] then Result:=I;
	end
else
	begin
	{Find closest vertex}
	for I:=0 to Lengths.ItemIndex-1 do
		if Lengths[I]<Lengths[Result] then Result:=I;
	end;
end;



procedure TSpanningHeap.Add(Parent,Child: integer; ParentVertex,ChildVertex: T3DVector; AdjCount: integer);
{Store adjacency information on heap}
var Length: double;
begin
Length:=VectorABS3D(VectorSubtract3D(ChildVertex,ParentVertex));
Lengths.Add(Length);
AdjCounts.Add(AdjCount);
ParentInx.Add(Parent);
ChildInx.Add(Child);
end;



constructor TSpanningHeap.Create;
{Create Spanning Heap object}
begin
inherited Create;
ParentInx:=THugeInteger.Create(5);
ChildInx:=THugeInteger.Create(5);
AdjCounts:=THugeInteger.Create(5);
Lengths:=THugeDouble.Create(5);
end;


destructor TSpanningHeap.Destroy;
{Destroy Spanning Heap Object}
begin
ParentInx.Free;
ChildInx.Free;
Lengths.Free;
AdjCounts.Free;
inherited Destroy;
end;

{-----------------------------------------------------------------}


procedure FindSpanningTree(Source,Destination: TSurveyGraph; CloseEdges: TLoopList; ByIncidence: boolean);
{Find spanning tree of Soure and store result in Destination}
{Stores edges that not in the spanning tree in CloseEdges (These signify Loops)}
{If "ByIncidence" is true the algorithm favors vertices with the most connections}
{ otherwise, it favors the shortest edges which means Prim's minimum spannign tree.}
var I,PInx,CInx: integer;
var Heap: TSpanningHeap;
var Edge: TEdge;

	function GetMaxIncidence: integer;
	{Return Vertex with maximum incidence}
	var I: integer;
	begin
	Result:=0;
	for I:=0 to Source.AdjacentVertices.ItemIndex-1 do
	 with Source do
	 if AdjacentVertices[I].ItemIndex>AdjacentVertices[Result].ItemIndex then Result:=I;
	end;


	procedure InsertAllAdjacent(Inx: integer);
	{Insert all vertices adjacent to Inx in heap}
	var I,Adj: integer;
	begin
	for I:=0 to Source.AdjacentVertices[Inx].ItemIndex-1 do
 	 with Source do
		begin
		Adj:=AdjacentVertices[Inx][I];
		if VertexValue[Adj]=0 then
		Heap.Add(Inx,Adj,Vertices[Inx],Vertices[Adj],AdjacentVertices[Adj].ItemIndex);
		end;
	end;


begin
{build and initialize objects}
Heap:=TSpanningHeap.Create;
Destination.Clear;
CloseEdges.Clear;
try
{Preset our graphs}
Source.SetColors(0);
CopyAllVertices(Source,Destination);

{If by incidence, start with vertex with max incidence}
if ByIncidence then I:=GetMaxIncidence else I:=0;

{Store initial vertex neighbor in heap}
InsertAllAdjacent(I);
Source.VertexValue[I]:=1;
while Heap.ChildInx.ItemIndex>0 do
	begin
	I:=Heap.GetBestHeapItem(ByIncidence);
	PInx:=Heap.ParentInx[I];
	CInx:=Heap.ChildInx[I];
	Heap.Delele(I);
	{Have we visited this vertex before?}
	if Source.VertexValue[CInx]=0 then
		begin
		{Unvisited edges are part of the spanning tree}
		Destination.AddEdge(Source.Names[PInx],Source.Names[CInx]);
		{Save neighbors on heap}
		InsertAllAdjacent(CInx);
		{Mark vertex as visted}
		Source.VertexValue[CInx]:=1;
		end
	else
		begin
		{Save previously visited as loop markers}
		Edge.Inx1:=Destination.Names.Search(Source.Names[PInx]);
		Edge.Inx2:=Destination.Names.Search(Source.Names[CInx]);
		TEdgeArray(CloseEdges.Add).Add(Edge);
		end;
	end;
{Clean up}
finally Heap.Free; end;
end;



{----------------------------------------------------------------------------}


function FindCommonParent(SpanTree: TSurveyGraph; Edge: TEdge): integer;
{ Locate The Common Node In Two Branches Of A Loop }
{ By Walking Backward Until You Find Common Node }
{ Returns}
var Back1,Back2,Start1,Start2,T1,T2,I: integer;
begin
Start1:=Edge.Inx1; Start2:=Edge.Inx2;
Back1:=Start1; Back2:=Start2;
while True do
	begin
	{ Backup One Station And Test One Side Of The Loop }
	T1:=SpanTree.Parent[Back1];
	if T1>-1 then
		begin
		Back1:=T1;
		I:=Start2;
		while True do
			begin
			if Back1=I then  begin Result:= Back1; exit; end;
			if I=Back2 then break;
			I:=SpanTree.Parent[I];
			end;
		end;

	{ Backup One Station And Test Other Side Of The Loop }
	T2:=SpanTree.Parent[Back2];
	if T2>-1 then
		begin
		Back2:=T2;
		I:=Start1;
		while True do
			begin
			if Back2=I then  begin Result:= Back2; exit; end;
			if I=Back1 then break;
			I:=SpanTree.Parent[I];
			end;
		end;
	if (T1<0) and (T2<0) then  begin Result:= T1; exit; end;
	end;
end;	{ Of Lcommon }



procedure StoreLoopSides(SpanTree: TSurveyGraph; EdgeArray: TEdgeArray; Inx,Common: integer);
{Store all edges on one side of loop, starting at Inx to Common}
var P: integer;
var Edge: TEdge;
begin
{Follow parents back to common vertex}
while Inx<>Common do
	begin
	{Get the parent vertex}
	P:=SpanTree.Parent[Inx];
	{Exit if no parent}
	if P<0 then exit;
	{Store edge}
	Edge.Inx1:=P; Edge.Inx2:=Inx;
	EdgeArray.Add(Edge);
	Inx:=P;
	end;
end;



procedure FindLoopEdges(SpanTree: TSurveyGraph; EdgeArray: TEdgeArray);
{Find and store loop from SpanTre into EdgeArray}
{The first Edge in EdgeArray will be the closing edge}
var ClosingEdge: TEdge;
begin
ClosingEdge:=EdgeArray.Edges[0];
EdgeArray.Common:=FindCommonParent(SpanTree,ClosingEdge);

StoreLoopSides(SpanTree,EdgeArray,ClosingEdge.Inx1,EdgeArray.Common);
EdgeArray.SecondHalf:=EdgeArray.ItemIndex;
StoreLoopSides(SpanTree,EdgeArray,ClosingEdge.Inx2,EdgeArray.Common);
end;


{================== Routines to optimize loops ===============================}



function CenterEdges(E: TEdgeArray; Graph: TSurveyGraph): T3DVector;
{Return center of gravity of all edges}
var I: integer;
begin
Result:=MakeVector3D(0,0,0);
for I:=0 to E.ItemIndex-1 do
 Result:=VectorAdd3D(Result,Graph.Vertices[E[I].Inx2]);
Result:=ScalarDivide3D(Result,E.ItemIndex);
end;



function GetAverageRadius(Loop: TEdgeArray; Graph: TSurveyGraph): double;
{Get the average radius from center point of loop}
var I: integer;
var C: T3DVector;
var Sum: double;
begin
C:=CenterEdges(Loop,Graph);
Sum:=0;
for I:=0 to Loop.ItemIndex-1 do
	Sum:=Sum +  VectorABS3D(VectorSubtract3D(Graph.Vertices[Loop[I].Inx2],C));
Result:=Sum;
end;

{------------- Routines to sort loops by perimeter and area ------------------}


var CompareGraph: TSurveyGraph;	{Current graph from which loops are sorted}

function CompareLoopSize(Item1,Item2: pointer): integer;
{Call back to compares for Quick Sort in HugeArray}
{Returns value so edges are sorted descending sizes}
type PEdgeArray= ^TEdgeArray;
var EA1,EA2: TEdgeArray;
begin
{Get underlying EdgeArrays from pointers}
EA1:=PEdgeArray(Item1)^;
EA2:=PEdgeArray(Item2)^;
{Return difference in edge length}
Result:=EA2.ItemIndex - EA1.ItemIndex;
{If loops are the same size test average radious}
if Result=0 then
	Result:=Trunc(GetAverageRadius(EA2,CompareGraph) - GetAverageRadius(EA1,CompareGraph));
end;


procedure SortLoopsBySize(Loops: TLoopList; Graph: TSurveyGraph);
{Routine to sort loop by size using generic Quick Sort built into HugeArray}
begin
CompareGraph:=Graph;
Loops.Sort(CompareLoopSize);
end;


{---------------------- Routines To Find Short cuts ---------------------------}

{FIFO object to store adjacent edges to a particular node}

type TFIFOList = class(TList)
 public
 procedure Insert(V: integer);
 function Pop: integer;
 end;


procedure TFIFOList.Insert(V: integer);
{Insert vertex index}
begin
Add(Pointer(V));
end;


function TFIFOList.Pop: integer;
{Simulate FIFO by getting first item in list and then deleting it}
{The delete might be expensive. Might be optimized using a circular buffer}
begin
Result:=0;
if Count>0 then
	begin
	Result:=integer(Items[0]);
	Delete(0);
	end;
end;



function ShortestDistance(Graph: TSurveyGraph; Common: TEdge; MaxDist: integer; Shortest: TEdgeArray): boolean;
{Find shortest distance between ends of "Common". }
{Length of new loop will be shorter than MaxDist}
var Queue: TFIFOList;
var Parents: THugeInteger;
var X,U,V,VDist: integer;
var I,VertexCount,EInx: integer;
var Edge: TEdge;
var Found: boolean;
begin
Found:=False;
VertexCount:=Graph.Vertices.ItemIndex;
{Create parent array}
Parents:=THugeInteger.Create(VertexCount);
{Create Queue}
Queue:=TFIFOList.Create;
try
{Create a parent for each vertex in the graph}
for I:=0 to VertexCount-1 do Parents.Add(-1);
{Vertex "Color" is used to exclude "common edges" that we've already used }
Graph.SetColors(0);

{Vertex Value is used to track the distance of each vertex}
{So, set distance of first vertex to one }
Graph.VertexValue[Common.Inx1]:=1;
{Insert first vertex of closing shot}
Queue.Insert(Common.Inx1);

while (Queue.Count>0) and not Found do
	begin
	{Get next vertex}
	U:=Queue.Pop;
	{Calculate current distance}
	VDist:=Graph.VertexValue[U] + 1;
	{Exit if there is no shorter path}
	if VDist>MaxDist then break;
	{Examine all adjacent vertices}
	for I:=0 to Graph.AdjacentVertices[U].ItemIndex-1 do
		begin
		{Get adjacent edge}
		V:=Graph.AdjacentVertices[U][I];
		Edge.Inx1:=U; Edge.Inx2:=V;
		{Get edge index from graph}
		EInx:=Graph.AdjacentEdges[U][I];
		{Ignore if it is the common edge or excluded}
		if not EqualEdges(Edge,Common) and (Graph.Edges[EInx].Color=0) then
			begin
			{if Distance is zero the vertex has not been visited}
			if Graph.VertexValue[V]=0 then
				begin
				{Set its parent, distance and insert in queue}
				Parents[V]:=U;
				Graph.VertexValue[V]:=VDist;
				Queue.Insert(V);
				{Other end of Common shot? then we have a short cut}
				if V = Common.Inx2 then begin Found:=True; break; end;
				end;
			end
		end;
	end;

{Gather edges for new, shorter loop}
if Found then
	begin
	{Start with the common edge}
	Shortest.Add(Common);
	{Backtrack over new path and gather edges}
	while Parents[V]<>-1 do
		begin
		Edge.Inx1:=V;
		V:=Parents[V];
		Edge.Inx2:=V;
		Shortest.Add(Edge);
		end;
	{Mark common edge so we don't use it again}
	EInx:=Graph.Edges.FindEdge(Common);
	Graph.Edges.Color[EInx]:=1;
	end;
finally
 Queue.Free;
 Parents.Free;
 end;
Result:=Found;
end;


{------------------------------------------------------------------------------}


procedure FindShortCuts(LoopList: TLoopList; Graph: TSurveyGraph; Optimize: boolean);
{Find short cut for each loop in the graph, if Optimize is true}
{Note: Optimize works at this level for debugging purposes}
{Because loops are still sorted if optimization is off }
var I: integer;
var ShortestLoop: TEdgeArray;
begin
ShortestLoop:=TEdgeArray.Create;
try
SortLoopsBySize(LoopList,Graph);
Graph.Edges.SetColors(0);
if Optimize then
for I:=0 to LoopList.ItemIndex-1 do
	begin
	ShortestLoop.Clear;
	if ShortestDistance(Graph,LoopList[I][0],LoopList[I].ItemIndex-2,ShortestLoop) then
		begin
		LoopList[I].Assign(ShortestLoop);
		end;
	end;
finally ShortestLoop.Free; end;
end;


{------------------------------------------------------------------------------}


procedure FindLoops(Source: TSurveyGraph; LoopList: TLoopList; Optimize: boolean);
{Find loops in "Source" graph. Loops are returned in "LoopList"}
{If Optimize is false, Loops are Fundamental cycles }
{If Optimize is True, Loops are minimum cycles}
var SpanTree: TSurveyGraph;
var I: integer;
begin
SpanTree:=TSurveyGraph.Create;
try
{Create a spanning tree from graph}
FindSpanningTree(Source,SpanTree,LoopList,True);
{Collect the edges for eah loop}
for I:=0 to LoopList.ItemIndex-1 do FindLoopEdges(SpanTree, LoopList[I]);
{Optimize by finding shortcut}
FindShortCuts(LoopList,Source,Optimize);
finally SpanTree.Free; end;
end;



end.
