unit CaveTypes;

interface

uses Controls,Math,CaveLib,Vector,DSupport;


{================== Plot File Data Structures ================}

{ Arrays Of Coordinates }

type TCaveDate = record
 Day, Month: byte;
 Year: word;
 end;


type TLocation = record
 case integer of
 0: (East,North,Vert: double);
 1: (X,Y,Z: double);
 2: (Vector: T3DVector);
 end;


type TPlotVector=record
 Command: char;
 Station: String[StationNameSize];
 Location: TLocation;
 Distance: double;
 Up,Down,Right,Left: single;
 end;

type PPlotVector=^TPlotVector;


type TSurveyType = (stSurvey,stFeature,stRange);
type TPlotAction = (paPlot,paExclude,paDepth,paPassage);

type TSurveyID=record
 Name: String[SurveyNameSize];
 Date: TCaveDate;
 Comment: string[SurveyCommentSize];
 SurveyType: TSurveyType;
 Action: TPlotAction;
 ShowStations: boolean;
 Color,Symbol: byte;
 Start,SegStart: integer;
 Min,Max: TLocation;
 end;

type PSurveyID=^TSurveyID;


type
 TCaveSection=record
 Name: String[SectionNameSize];
 Color: byte;
 Action: TPlotAction;
 ShowStations: boolean;
 Start: longint;
 end;


type TBoundBox = record
 Min,Max: TLocation;
 end;

type TBoxCorner = (TopNorthWest,TopNorthEast,TopSouthEast,TopSouthWest,
		    BottomNorthWest,BottomNorthEast,BottomSouthEast,BottomSouthWest);



{================== Raw Data File Structures ====================}

{ Shot data structures }

type TShotMeasure = record
 Bearing,Length,Inclination: double;
 end;

type TPassDimensions = record
 Left,Right,Up,Down: double;
 end;

type TStation = record
 Name: string[StationNameSize];
 Dimensions: TPassDimensions;
 end;

type TShotFlag = (sfExclude,sfNoPlot,sfNoLength,sfNoClose);

type TShotFlags = set of TShotFlag;


const MaxUserFlags = 20;

type TShot = record
 FromStation,ToStation: TStation;
 ForeShot, BackShot: TShotMeasure;
 Comment: string[ShotCommentSize];
 Flags: TShotFlags;
 UserFlags: string[MaxUserFlags];
 end;


{ Survey data structures }

{Entry Order items}

type TMeasureItems = (miBearing,miLength,miInclination);
type TDimensionItems = (diLeft,diRight,diUp,diDown);

{Units definitions}

type TLengthUnits = (luFeet,luFeetIn,luMeters);
type TCompassUnits = (cuDegrees,cuQuads,cuGrads);
type TInclinationUnits = (iuDegrees,iuDegMinutes,iuGrade,iuGrads,iuDepth);

type TInstrumentID = record
 Correction: double;
 Owner: string[40];
 SerialNumber: string[40];
 end;

type TInstrumentSet = record
 Compass,Tape,Inclinometer: TInstrumentID;
 end;

type TTeamMember = record
 FirstName,LastName: string[20];
 Job: string[40];
 end;

const MaxTeamMembers = 8;

type TSurveyTeam = record
 Count: integer;
 Members: array [0..MaxTeamMembers-1] of TTeamMember;
 Comment: shortstring;
 end;

const MeasureOrderCount = 3;
const DimOrderCount = 4;

type TSurveyHeader = record
 CaveName: string[CaveNameSize];
 SurveyName: string[SurveyNameSize];
 SurveyComment: string[SurveyCommentSize];
 Instrument1,Instrument2: TInstrumentSet;
 DiscoveryDate,SurveyDate: TDate;
 SurveyTeam: TSurveyTeam;
 Declination: double;

 MeasureOrder: array [0..MeasureOrderCount-1] of TMeasureItems;
 DimensionOrder: array [0..DimOrderCount-1] of TDimensionItems;
 TapeUnits: TLengthUnits;
 CompassUnits: TCompassUnits;
 InclinationUnits: TInclinationUnits;
 LRUDUnits: TLengthUnits;
 BackSight: boolean;
 end;

{File-level structures}

type TCaveFileHeader = record
  SectionName: shortstring;
  Comment: string[SurveyCommentSize];
  end;

{ Fixed Station Structures }

type TFixedStation = record
 Name: string[StationNameSize];
 Location: T3DVector;
 Zone: integer;
 end;

{ Project Level Structures}

type TProjectItem = record
 Filename: shortstring;
 UseMeters: boolean;
 end;

{Support routines}

function GetSurveyTypeStr(SurveyType: TSurveyType; Short: boolean): string;
procedure VectorToPolar3D(P1,P2: T3DVector; var Azimuth,Inclination,Length: double);
function GetBoxCorner(Box: TBoundBox; Corner: TBoxCorner): T3DVector;
function MakeBox(Min,Max: TLocation): TBoundBox;
function MakeShotMeasure(Azm,Len,Slope: double): TShotMeasure;
function MakePassDimension(Left,Right,Up,Down: double): TPassDimensions;


implementation


{--------------- Plot Support Routines ----------------------------------------}


function GetSurveyTypeStr(SurveyType: TSurveyType; Short: boolean): string;
{Get label string for survey type }
begin
case SurveyType of
 stSurvey: if Short then Result:='S' else Result:='Survey';
 stFeature: if Short then  Result:='F' else Result:='Feature';
 stRange: if Short then Result:='R' else Result:='Range';
 end;
end;




procedure VectorToPolar3D(P1,P2: T3DVector; var Azimuth,Inclination,Length: double);
{ Convert two points, from P1 to P2, to Azimuth and Inclination }
var D:T3DVector;
begin
D:=VectorSubtract3D(P2,P1);
Length:=VectorABS3D(D);
Azimuth:=(ArcTan2(D.X,D.Y) * Radian2Degree);
if Azimuth<0 then Azimuth:=360 + Azimuth;

if D.Z=0 then Inclination:=0
else
	begin
	Inclination:=D.Z / Length;
	if Inclination>1 then Inclination:=1;
	if Inclination<-1 then Inclination:=-1;
	Inclination:=ArcSin(Inclination)  * Radian2Degree;
	end;
end;


function GetBoxCorner(Box: TBoundBox; Corner: TBoxCorner): T3DVector;
{Choose one corner of a box}
begin
case Corner of
 TopNorthWest: Result:=MakeVector3D(Box.Min.East,Box.Max.North,Box.Max.Vert);
 TopNorthEast: Result:=MakeVector3D(Box.Max.East,Box.Max.North,Box.Max.Vert);
 TopSouthEast: Result:=MakeVector3D(Box.Max.East,Box.Min.North,Box.Max.Vert);
 TopSouthWest: Result:=MakeVector3D(Box.Min.East,Box.Min.North,Box.Max.Vert);
 BottomNorthWest: Result:=MakeVector3D(Box.Min.East,Box.Max.North,Box.Min.Vert);
 BottomNorthEast: Result:=MakeVector3D(Box.Max.East,Box.Max.North,Box.Min.Vert);
 BottomSouthEast: Result:=MakeVector3D(Box.Max.East,Box.Min.North,Box.Min.Vert);
 BottomSouthWest: Result:=MakeVector3D(Box.Min.East,Box.Min.North,Box.Min.Vert);
 end;
end;


function MakeBox(Min,Max: TLocation): TBoundBox;
begin
Result.Min:=Min;
Result.Max:=Max;
end;


function MakeShotMeasure(Azm,Len,Slope: double): TShotMeasure;
begin
Result.Bearing:=Azm;
Result.Length:=Len;
Result.Inclination:=Slope;
end;


function MakePassDimension(Left,Right,Up,Down: double): TPassDimensions;
begin
Result.Left:=Left;
Result.Right:=Right;
Result.Up:=Up;
Result.Down:=Down;
end;


end.
