unit Hash;

{
This unit is used to create a HashTable object. The object
has the following capabilities:

1. It is based on the THugeString object so it can maintain
a list of strings that automatically grows as items are added.

2. The strings added to the list are hashed so the list can be
searched with maximum efficiency.
}

interface

uses SysUtils,Math,HugeArry;

{------------------------------------------------------------}
{----------- Declaration for Hash Table object --------------}
{------------------------------------------------------------}


{ Variables For Hash Table Lookup }

const HashEmpty=-1;		{ Empty Symbol Pointer }

{Hashtable object Declaration}

type THashTable = class(TObject)
 private
  FBoxPower: integer;			{ Power of the number of boxes}
  FBoxCount: integer;			{ Number Of Boxes MUST be power of two}
  FHashMask: integer;			{ Hash Mask }
  HashLink: THugeLongint;		{ List Linkage Pointers }
  HashBox: THugeLongint;		{ Hash Boxes (symbol List Headers) }
  HashValue: Integer;			{ Current Identifier Hash Code }
  FCaseSensitive: boolean;
  procedure SetCaseSensitive(V: boolean);
  procedure SetBoxPower(V: integer);
  procedure SetupHashBoxes;
  function ReadString(I: integer): shortstring;
  procedure WriteString(I: integer; V: shortstring);
  function GetCount: integer;
 protected
  procedure ClearHashBox;
  function GetHashValue(S: ShortString): integer;
 public
  Strings: THugeString;		{ String table }
  constructor Create(StartAlloc,StrSize: LongInt);
  destructor Destroy; override;
  function Search(S: ShortString): Longint;
  function Insert(S: ShortString): integer;
  function InsertPointer(S: shortString; Ptr: integer): integer;
  procedure Rehash;
  procedure GetDistribution(var Dist: THugeInteger);
  procedure Clear;
  property Value[Inx: integer]: shortstring read ReadString write WriteString; default;
  property Count: integer read GetCount;
  property BoxPower: integer read FBoxPower write SetBoxPower;
  property CaseSensitive: boolean read FCaseSensitive write SetCaseSensitive;
 end;

const FNV_prime = 16777619;		{Magic prime number which depends on hash size}
const offset_basis = 2166136261;	{Magic number used as starting hash also depends on hash size}


 implementation


{-------------------------------------------------------------------------}
{                  Implementation for hash table object                   }
{-------------------------------------------------------------------------}


procedure THashTable.SetupHashBoxes;
{Create and intialize hash boxes}
begin
HashBox.ReAllocMem(FBoxCount);
HashBox.ItemIndex:=FBoxCount;
ClearHashBox;
Rehash;
end;



procedure THashTable.SetBoxPower(V: integer);
{Set number of hash boxes based on power of two}
begin
if V<>FBoxPower then
	begin
	if V<4 then FBoxPower:=4 else FBoxPower:=V;
	FBoxCount:=1 shl FBoxPower;
	FHashMask:=FBoxCount-1;
	SetupHashBoxes;
	end;
end;




function THashTable.GetHashValue(S: shortString): integer;
{ Return Hash For String S }
{ Hashing algorith based on the FNV-1as (Fowler / Noll / Vo)}
var P,T : Integer;
begin
if not CaseSensitive then S:=LowerCase(S);
Result:=offset_basis;
for P:=1 to Length(S) do
	begin
	Result:=Result xor byte(S[P]);
	Result:= Result * FNV_Prime;
	end;
Result:=((Result shr FBoxPower) xor Result) and  FHashMask;
end;




procedure  THashTable.ClearHashBox;
{ Clear Hash Box }
var P : Integer;
begin
for P:=0 to FBoxCount-1 do  HashBox.Value[P]:=HashEmpty;
end;



procedure THashTable.Clear;
{Clear all the hash boxes and reinitialize}
begin
Strings.Clear;
HashLink.Clear;
SetupHashBoxes;
end;



procedure  THashTable.Rehash;
{ Re-do The Hash Links For The Table }
var P : Longint;
begin
ClearHashBox;
HashLink.ItemIndex:=Strings.ItemIndex;
for P:=0 to Strings.ItemIndex-1 do
	begin
	HashValue:=GetHashValue(Strings.Value[P]);
	HashLink.Value[P]:=HashBox.Value[HashValue];
	HashBox.Value[HashValue]:=P;
	end;
end;


function THashTable.Search(S: shortString): Longint;
{ Search For Station S, return index of string }
{ Returns -1 if not found }
begin
HashValue:=GetHashValue(S);
Result:=HashBox.Value[HashValue];
while True do
	begin
	if Result=HashEmpty then break;
	if Strings.Compare(S,Result) then break;
	Result:=HashLink.Value[Result];
	end;
end;


procedure THashTable.GetDistribution(var Dist: THugeInteger);
{Get the distribution of hashbox values - for testing purposes}
var I,Inx,Cnt: integer;
begin
Dist.Clear;
for I:=0 to HashBox.ItemIndex-1 do
	begin
	Inx:=HashBox[I];
	Cnt:=0;
	while Inx<>HashEmpty do
		begin
		Inc(Cnt);
		Inx:=HashLink.Value[Inx];
		end;
	Dist.Add(Cnt)
	end;
end;


procedure THashTable.SetCaseSensitive(V: boolean);
{Used to set the Case Sensitivity property}
begin
if V<>FCaseSensitive then
	begin
	FCaseSensitive:=V;
	Strings.CaseSensitive:=V;
	Rehash;
	end;
end;


{-------------------------------------------------------------------}


constructor THashTable.Create(StartAlloc,StrSize: LongInt);
{Creation routine for hash table object }
begin
inherited Create;
Strings:=THugeString.Create(StartAlloc,StrSize);
HashLink:=THugeLongInt.Create(StartAlloc);
HashBox:=THugeLongInt.Create(5);
CaseSensitive:=True;
BoxPower:=11;	{2048}
end;


destructor THashTable.Destroy;
{Destroy hash table object }
begin
Strings.Free;
HashLink.Free;
HashBox.Free;
inherited Destroy;
end;



function  THashTable.GetCount: integer;
begin
Result:=Strings.ItemIndex;
end;


function THashTable.ReadString(I: integer): shortstring;
{Property function so hash table can use array-type syntax}
begin
Result:=Strings[I];
end;


procedure THashTable.WriteString(I: integer; V: shortstring);
{Property function so hash table can use array-type syntax}
begin
Strings[I]:=V;
end;




function THashTable.InsertPointer(S: shortString; Ptr: integer): integer;
{ Insert string and pointer Into Hash Table}
{ Pointer can point to any object the user }
{wishes to associate with a specific string }
begin
HashValue:=GetHashValue(S);
Result:=Strings.AddPointer(S,Ptr);
HashLink.Add(HashBox.Value[HashValue]);
HashBox.Value[HashValue]:=HashLink.ItemIndex-1;
end;



function THashTable.Insert(S: shortString): integer;
{ Insert string Into Hash Table }
begin
Result:=InsertPointer(S,0);
end;


end.
